@@ -9,14 +9,18 @@ from .clipboard import ClipboardData
 from .completion import Completer, Completion, CompleteEvent
 from .document import Document
 from .enums import IncrementalSearchDirection
-from .filters import Never, to_simple_filter
+from .filters import to_simple_filter
 from .history import History, InMemoryHistory
 from .search_state import SearchState
 from .selection import SelectionType, SelectionState
-from .utils import Callback
+from .utils import Event
+from .cache import FastDictCache
 from .validation import ValidationError
 
+from six.moves import range
+
 import os
+import re
 import six
 import subprocess
 import tempfile
@@ -27,10 +31,12 @@ __all__ = (
     'Buffer',
     'indent',
     'unindent',
+    'reshape_text',
 )
 
+
 class EditReadOnlyBuffer(Exception):
-    " Attempt editing of read-only buffer. "
+    " Attempt editing of read-only :class:`.Buffer`. "
 
 
 class AcceptAction(object):
@@ -38,10 +44,10 @@ class AcceptAction(object):
     What to do when the input is accepted by the user.
     (When Enter was pressed in the command line.)
 
-    :param handler: (optional) A callable which accepts a CLI and `Document'
-                    that is called when the user accepts input.
-    :param render_cli_done: When using a handler, first render the CLI in the
-                    'done' state, then call the handler. This
+    :param handler: (optional) A callable which takes a
+        :class:`~prompt_toolkit.interface.CommandLineInterface` and
+        :class:`~prompt_toolkit.document.Document`. It is called when the user
+        accepts input.
     """
     def __init__(self, handler=None):
         assert handler is None or callable(handler)
@@ -50,7 +56,8 @@ class AcceptAction(object):
     @classmethod
     def run_in_terminal(cls, handler, render_cli_done=False):
         """
-        Create an `AcceptAction` that runs the given handler in the terminal.
+        Create an :class:`.AcceptAction` that runs the given handler in the
+        terminal.
 
         :param render_cli_done: When True, render the interface in the 'Done'
                 state first, then execute the function. If False, erase the
@@ -109,7 +116,7 @@ class CompletionState(object):
 
     def go_to_index(self, index):
         """
-        Create a new CompletionState object with the new index.
+        Create a new :class:`.CompletionState` object with the new index.
         """
         return CompletionState(self.original_document, self.current_completions, complete_index=index)
 
@@ -149,32 +156,38 @@ class Buffer(object):
     current input line and implements all text manupulations on top of it. It
     also implements the history, undo stack and the completion state.
 
-    :attr completer : :class:`~prompt_toolkit.completion.Completer` instance.
-    :attr history: :class:`~prompt_toolkit.history.History` instance.
-    :attr callbacks: :class:`~.Callbacks` instance.
-
-    :attr tempfile_suffix: Suffix to be appended to the tempfile for the 'open
+    :param completer: :class:`~prompt_toolkit.completion.Completer` instance.
+    :param history: :class:`~prompt_toolkit.history.History` instance.
+    :param tempfile_suffix: Suffix to be appended to the tempfile for the 'open
                            in editor' function.
-    :attr is_multiline: SimpleFilter to indicate whether we should consider
-                        this buffer a multiline input. If so, key bindings can
-                        decide to insert newlines when pressing [Enter].
-                        (Instead of accepting the input.)
-    :param complete_while_typing: Filter instance. Decide whether or not to do
-                                  asynchronous autocompleting while typing.
 
-    :param on_text_changed: Callback instance or None.
-    :param on_text_insert: Callback instance or None.
-    :param on_cursor_position_changed: Callback instance or None.
-    :param enable_history_search: SimpleFilter to indicate when up-arrow partial
-        string matching is enabled. It is adviced to not enable this at the
-        same time as `complete_while_typing`, because when there is an
-        autocompletion found, the up arrows usually browse through the
-        completions, rather than through the history.
+    Events:
+
+    :param on_text_changed: When the buffer text changes. (Callable on None.)
+    :param on_text_insert: When new text is inserted. (Callable on None.)
+    :param on_cursor_position_changed: When the cursor moves. (Callable on None.)
+
+    Filters:
+
+    :param is_multiline: :class:`~prompt_toolkit.filters.SimpleFilter` to
+        indicate whether we should consider this buffer a multiline input. If
+        so, key bindings can decide to insert newlines when pressing [Enter].
+        (Instead of accepting the input.)
+    :param complete_while_typing: :class:`~prompt_toolkit.filters.SimpleFilter`
+        instance. Decide whether or not to do asynchronous autocompleting while
+        typing.
+    :param enable_history_search: :class:`~prompt_toolkit.filters.SimpleFilter`
+        to indicate when up-arrow partial string matching is enabled. It is
+        adviced to not enable this at the same time as `complete_while_typing`,
+        because when there is an autocompletion found, the up arrows usually
+        browse through the completions, rather than through the history.
+    :param read_only: :class:`~prompt_toolkit.filters.SimpleFilter`. When True,
+        changes will not be allowed.
     """
     def __init__(self, completer=None, auto_suggest=None, history=None,
                  validator=None, tempfile_suffix='',
-                 is_multiline=Never(), complete_while_typing=Never(),
-                 enable_history_search=Never(), initial_document=None,
+                 is_multiline=False, complete_while_typing=False,
+                 enable_history_search=False, initial_document=None,
                  accept_action=AcceptAction.IGNORE, read_only=False,
                  on_text_changed=None, on_text_insert=None, on_cursor_position_changed=None):
 
@@ -188,9 +201,9 @@ class Buffer(object):
         assert completer is None or isinstance(completer, Completer)
         assert auto_suggest is None or isinstance(auto_suggest, AutoSuggest)
         assert history is None or isinstance(history, History)
-        assert on_text_changed is None or isinstance(on_text_changed, Callback)
-        assert on_text_insert is None or isinstance(on_text_insert, Callback)
-        assert on_cursor_position_changed is None or isinstance(on_cursor_position_changed, Callback)
+        assert on_text_changed is None or callable(on_text_changed)
+        assert on_text_insert is None or callable(on_text_insert)
+        assert on_cursor_position_changed is None or callable(on_cursor_position_changed)
 
         self.completer = completer
         self.auto_suggest = auto_suggest
@@ -204,6 +217,9 @@ class Buffer(object):
         self.enable_history_search = enable_history_search
         self.read_only = read_only
 
+        # Text width. (For wrapping, used by the Vi 'gq' operator.)
+        self.text_width = 0
+
         #: The command buffer history.
         # Note that we shouldn't use a lazy 'or' here. bool(history) could be
         # False when empty.
@@ -212,9 +228,12 @@ class Buffer(object):
         self.__cursor_position = 0
 
         # Events
-        self.on_text_changed = on_text_changed or Callback()
-        self.on_text_insert = on_text_insert or Callback()
-        self.on_cursor_position_changed = on_cursor_position_changed or Callback()
+        self.on_text_changed = Event(self, on_text_changed)
+        self.on_text_insert = Event(self, on_text_insert)
+        self.on_cursor_position_changed = Event(self, on_cursor_position_changed)
+
+        # Document cache. (Avoid creating new Document instances.)
+        self._document_cache = FastDictCache(Document, size=10)
 
         self.reset(initial_document=initial_document)
 
@@ -237,6 +256,9 @@ class Buffer(object):
         # State of the selection.
         self.selection_state = None
 
+        # When doing consecutive up/down movements, prefer to stay at this column.
+        self.preferred_column = None
+
         # State of complete browser
         self.complete_state = None  # For interactive completion through Ctrl-N/Ctrl-P.
 
@@ -264,10 +286,23 @@ class Buffer(object):
 
     def _set_text(self, value):
         """ set text at current working_index. Return whether it changed. """
-        original_value = self._working_lines[self.working_index]
-        self._working_lines[self.working_index] = value
+        working_index = self.working_index
+        working_lines = self._working_lines
 
-        return value != original_value
+        original_value = working_lines[working_index]
+        working_lines[working_index] = value
+
+        # Return True when this text has been changed.
+        if len(value) != len(original_value):
+            # For Python 2, it seems that when two strings have a different
+            # length and one is a prefix of the other, Python still scans
+            # character by character to see whether the strings are different.
+            # (Some benchmarking showed significant differences for big
+            # documents. >100,000 of lines.)
+            return True
+        elif value != original_value:
+            return True
+        return False
 
     def _set_cursor_position(self, value):
         """ Set cursor position. Return whether it changed. """
@@ -335,6 +370,7 @@ class Buffer(object):
         self.complete_state = None
         self.selection_state = None
         self.suggestion = None
+        self.preferred_column = None
 
         # fire 'on_text_changed' event.
         self.on_text_changed.fire()
@@ -344,6 +380,10 @@ class Buffer(object):
         self.validation_error = None
         self.complete_state = None
 
+        # Unset preferred_column. (Will be set after the cursor movement, if
+        # required.)
+        self.preferred_column = None
+
         # Note that the cursor position can change if we have a selection the
         # new position of the cursor determines the end of the selection.
 
@@ -353,15 +393,16 @@ class Buffer(object):
     @property
     def document(self):
         """
-        Return :class:`Document` instance from the current text and cursor
-        position.
+        Return :class:`~prompt_toolkit.document.Document` instance from the
+        current text, cursor position and selection state.
         """
-        return Document(self.text, self.cursor_position, selection=self.selection_state)
+        return self._document_cache[
+            self.text, self.cursor_position, self.selection_state]
 
     @document.setter
     def document(self, value):
         """
-        Set :class:`Document` instance.
+        Set :class:`~prompt_toolkit.document.Document` instance.
 
         This will set both the text and cursor position at the same time, but
         atomically. (Change events will be triggered only after both have been set.)
@@ -370,11 +411,12 @@ class Buffer(object):
 
     def set_document(self, value, bypass_readonly=False):
         """
-        Set :class:`Document` instance. Like the `document` property, but
-        accept an `bypass_readonly` argument.
+        Set :class:`~prompt_toolkit.document.Document` instance. Like the
+        ``document`` property, but accept an ``bypass_readonly`` argument.
 
-        :param bypass_readonly: When True, don't raise an `EditReadOnlyBuffer`
-                                exception, even when the buffer is read-only.
+        :param bypass_readonly: When True, don't raise an
+                                :class:`.EditReadOnlyBuffer` exception, even
+                                when the buffer is read-only.
         """
         assert isinstance(value, Document)
 
@@ -440,14 +482,28 @@ class Buffer(object):
 
         return '\n'.join(lines)
 
+    def transform_current_line(self, transform_callback):
+        """
+        Apply the given transformation function to the current line.
+
+        :param transform_callback: callable that takes a string and return a new string.
+        """
+        document = self.document
+        a = document.cursor_position + document.get_start_of_line_position()
+        b = document.cursor_position + document.get_end_of_line_position()
+        self.text = (
+            document.text[:a] +
+            transform_callback(document.text[a:b]) +
+            document.text[b:])
+
     def transform_region(self, from_, to, transform_callback):
         """
         Transform a part of the input string.
 
-        :param :from_: (int) start position.
-        :param :to: (int) end position.
-        :param :transform_callback: Callable which accepts a string and returns
-                                    the transformed string.
+        :param from_: (int) start position.
+        :param to: (int) end position.
+        :param transform_callback: Callable which accepts a string and returns
+            the transformed string.
         """
         assert from_ < to
 
@@ -465,13 +521,23 @@ class Buffer(object):
 
     def cursor_up(self, count=1):
         """ (for multiline edit). Move cursor to the previous line.  """
-        self.cursor_position += self.document.get_cursor_up_position(count=count)
+        original_column = self.preferred_column or self.document.cursor_position_col
+        self.cursor_position += self.document.get_cursor_up_position(
+            count=count, preferred_column=original_column)
+
+        # Remember the original column for the next up/down movement.
+        self.preferred_column = original_column
 
     def cursor_down(self, count=1):
         """ (for multiline edit). Move cursor to the next line.  """
-        self.cursor_position += self.document.get_cursor_down_position(count=count)
+        original_column = self.preferred_column or self.document.cursor_position_col
+        self.cursor_position += self.document.get_cursor_down_position(
+            count=count, preferred_column=original_column)
 
-    def auto_up(self, count=1):
+        # Remember the original column for the next up/down movement.
+        self.preferred_column = original_column
+
+    def auto_up(self, count=1, go_to_start_of_line_if_history_changes=False):
         """
         If we're not on the first line (of a multiline input) go a line up,
         otherwise go back in history. (If nothing is selected.)
@@ -479,11 +545,15 @@ class Buffer(object):
         if self.complete_state:
             self.complete_previous(count=count)
         elif self.document.cursor_position_row > 0:
-            self.cursor_position += self.document.get_cursor_up_position(count=count)
+            self.cursor_up(count=count)
         elif not self.selection_state:
             self.history_backward(count=count)
 
-    def auto_down(self, count=1):
+            # Go to the start of the line?
+            if go_to_start_of_line_if_history_changes:
+                self.cursor_position += self.document.get_start_of_line_position()
+
+    def auto_down(self, count=1, go_to_start_of_line_if_history_changes=False):
         """
         If we're not on the last line (of a multiline input) go a line down,
         otherwise go forward in history. (If nothing is selected.)
@@ -491,13 +561,18 @@ class Buffer(object):
         if self.complete_state:
             self.complete_next(count=count)
         elif self.document.cursor_position_row < self.document.line_count - 1:
-            self.cursor_position += self.document.get_cursor_down_position(count=count)
+            self.cursor_down(count=count)
         elif not self.selection_state:
             self.history_forward(count=count)
 
+            # Go to the start of the line?
+            if go_to_start_of_line_if_history_changes:
+                self.cursor_position += self.document.get_start_of_line_position()
+
     def delete_before_cursor(self, count=1):
         """
-        Delete character before cursor, return deleted character.
+        Delete specified number of characters before cursor and return the
+        deleted text.
         """
         assert count >= 0
         deleted = ''
@@ -515,7 +590,7 @@ class Buffer(object):
 
     def delete(self, count=1):
         """
-        Delete one character. Return deleted character.
+        Delete specified number of characters and Return the deleted text.
         """
         if self.cursor_position < len(self.text):
             deleted = self.document.text_after_cursor[:count]
@@ -525,7 +600,7 @@ class Buffer(object):
         else:
             return ''
 
-    def join_next_line(self):
+    def join_next_line(self, separator=' '):
         """
         Join the next line to the current one by deleting the line ending after
         the current line.
@@ -535,24 +610,24 @@ class Buffer(object):
             self.delete()
 
             # Remove spaces.
-            self.text = (self.document.text_before_cursor + ' ' +
+            self.text = (self.document.text_before_cursor + separator +
                          self.document.text_after_cursor.lstrip(' '))
 
-    def join_selected_lines(self):
+    def join_selected_lines(self, separator=' '):
         """
         Join the selected lines.
         """
         assert self.selection_state
 
         # Get lines.
-        from_, to = self.document.selection_range()
+        from_, to = sorted([self.cursor_position, self.selection_state.original_cursor_position])
 
         before = self.text[:from_]
         lines = self.text[from_:to].splitlines()
         after = self.text[to:]
 
         # Replace leading spaces with just one space.
-        lines = [l.lstrip(' ') + ' ' for l in lines]
+        lines = [l.lstrip(' ') + separator for l in lines]
 
         # Set new document.
         self.document = Document(text=before + ''.join(lines) + after,
@@ -739,7 +814,6 @@ class Buffer(object):
         Move forwards through the history.
 
         :param count: Amount of items to move forward.
-        :param history_search: When True, filter history using self.history_search_text.
         """
         self._set_history_search()
 
@@ -788,32 +862,28 @@ class Buffer(object):
 
     def copy_selection(self, _cut=False):
         """
-        Copy selected text and return :class:`ClipboardData` instance.
+        Copy selected text and return :class:`.ClipboardData` instance.
         """
-        if self.selection_state:
-            type = self.selection_state.type
+        new_document, clipboard_data = self.document.cut_selection()
+        if _cut:
+            self.document = new_document
 
-            # Take start and end of selection
-            from_, to = self.document.selection_range()
-
-            copied_text = self.text[from_:to]
-
-            # If cutting, remove the text and set the new cursor position.
-            if _cut:
-                self.document = Document(text=self.text[:from_] + self.text[to + 1:],
-                                         cursor_position=min(from_, to))
-
-            self.selection_state = None
-            return ClipboardData(copied_text, type)
-        else:
-            return ClipboardData('')
+        self.selection_state = None
+        return clipboard_data
 
     def cut_selection(self):
         """
-        Delete selected text and return :class:`ClipboardData` instance.
+        Delete selected text and return :class:`.ClipboardData` instance.
         """
         return self.copy_selection(_cut=True)
 
+    def paste_clipboard_data(self, data, before=False, count=1):
+        """
+        Insert the data from the clipboard.
+        """
+        assert isinstance(data, ClipboardData)
+        self.document = self.document.paste_clipboard_data(data, before=before, count=count)
+
     def newline(self, copy_margin=True):
         """
         Insert a line ending at the current position.
@@ -855,16 +925,21 @@ class Buffer(object):
         :param fire_event: Fire `on_text_insert` event. This is mainly used to
             trigger autocompletion while typing.
         """
+        # Original text & cursor position.
+        otext = self.text
+        ocpos = self.cursor_position
+
         # In insert/text mode.
         if overwrite:
-            # Don't overwrite the newline itself. Just before the line ending, it should act like insert mode.
-            overwritten_text = self.text[self.cursor_position:self.cursor_position+len(data)]
+            # Don't overwrite the newline itself. Just before the line ending,
+            # it should act like insert mode.
+            overwritten_text = otext[ocpos:ocpos + len(data)]
             if '\n' in overwritten_text:
                 overwritten_text = overwritten_text[:overwritten_text.find('\n')]
 
-            self.text = self.text[:self.cursor_position] + data + self.text[self.cursor_position+len(overwritten_text):]
+            self.text = otext[:ocpos] + data + otext[ocpos + len(overwritten_text):]
         else:
-            self.text = self.text[:self.cursor_position] + data + self.text[self.cursor_position:]
+            self.text = otext[:ocpos] + data + otext[ocpos:]
 
         if move_cursor:
             self.cursor_position += len(data)
@@ -873,31 +948,6 @@ class Buffer(object):
         if fire_event:
             self.on_text_insert.fire()
 
-    def paste_clipboard_data(self, data, before=False, count=1):
-        """
-        Insert the data from the clipboard.
-        """
-        assert isinstance(data, ClipboardData)
-
-        if data.type == SelectionType.CHARACTERS:
-            if before:
-                self.insert_text(data.text * count)
-            else:
-                self.cursor_right()
-                self.insert_text(data.text * count, fire_event=False)
-                self.cursor_left()
-
-        elif data.type == SelectionType.LINES:
-            if before:
-                self.cursor_position += self.document.get_start_of_line_position(after_whitespace=False)
-                self.insert_text((data.text + '\n') * count, move_cursor=False)
-            else:
-                self.cursor_position += self.document.get_end_of_line_position()
-                self.insert_text(('\n' + data.text) * count, move_cursor=False, fire_event=False)
-                self.cursor_down()
-
-            self.cursor_position += self.document.get_start_of_line_position(after_whitespace=True)
-
     def undo(self):
         # Pop from the undo-stack until we find a text that if different from
         # the current text. (The current logic of `save_to_undo_stack` will
@@ -935,7 +985,7 @@ class Buffer(object):
                 self.validator.validate(self.document)
             except ValidationError as e:
                 # Set cursor position (don't allow invalid values.)
-                cursor_position = e.index
+                cursor_position = e.cursor_position
                 self.cursor_position = min(max(0, cursor_position), len(self.text))
 
                 self.validation_error = e
@@ -1017,10 +1067,10 @@ class Buffer(object):
         # Do 'count' search iterations.
         working_index = self.working_index
         document = self.document
-        for i in range(count):
+        for _ in range(count):
             result = search_once(working_index, document)
             if result is None:
-                return # Nothing found.
+                return  # Nothing found.
             else:
                 working_index, document = result
 
@@ -1028,8 +1078,8 @@ class Buffer(object):
 
     def document_for_search(self, search_state):
         """
-        Return a `Document` instance that has the text/cursor position for this
-        search, if we would apply it.
+        Return a :class:`~prompt_toolkit.document.Document` instance that has
+        the text/cursor position for this search, if we would apply it.
         """
         search_result = self._search(search_state, include_current_position=True)
 
@@ -1041,11 +1091,11 @@ class Buffer(object):
 
     def apply_search(self, search_state, include_current_position=True, count=1):
         """
-        Return a `Document` instance that has the text/cursor position for this
-        search, if we would apply it.
+        Apply search. If something is found, set `working_index` and
+        `cursor_position`.
         """
-        search_result = self._search(search_state,
-            include_current_position=include_current_position, count=count)
+        search_result = self._search(
+            search_state, include_current_position=include_current_position, count=count)
 
         if search_result is not None:
             working_index, cursor_position = search_result
@@ -1059,7 +1109,8 @@ class Buffer(object):
         """
         Open code in editor.
 
-        :param cli: `CommandLineInterface` instance.
+        :param cli: :class:`~prompt_toolkit.interface.CommandLineInterface`
+            instance.
         """
         if self.read_only():
             raise EditReadOnlyBuffer()
@@ -1128,7 +1179,7 @@ class Buffer(object):
 
 def indent(buffer, from_row, to_row, count=1):
     """
-    Indent text of the `Buffer` object.
+    Indent text of a :class:`.Buffer` object.
     """
     current_row = buffer.document.cursor_position_row
     line_range = range(from_row, to_row)
@@ -1145,7 +1196,7 @@ def indent(buffer, from_row, to_row, count=1):
 
 def unindent(buffer, from_row, to_row, count=1):
     """
-    Unindent text of the `Buffer` object.
+    Unindent text of a :class:`.Buffer` object.
     """
     current_row = buffer.document.cursor_position_row
     line_range = range(from_row, to_row)
@@ -1165,3 +1216,48 @@ def unindent(buffer, from_row, to_row, count=1):
 
     # Go to the start of the line.
     buffer.cursor_position += buffer.document.get_start_of_line_position(after_whitespace=True)
+
+
+def reshape_text(buffer, from_row, to_row):
+    """
+    Reformat text, taking the width into account.
+    `to_row` is included.
+    (Vi 'gq' operator.)
+    """
+    lines = buffer.text.splitlines(True)
+    lines_before = lines[:from_row]
+    lines_after = lines[to_row + 1:]
+    lines_to_reformat = lines[from_row:to_row + 1]
+
+    if lines_to_reformat:
+        # Take indentation from the first line.
+        length = re.search(r'^\s*', lines_to_reformat[0]).end()
+        indent = lines_to_reformat[0][:length].replace('\n', '')
+
+        # Now, take all the 'words' from the lines to be reshaped.
+        words = ''.join(lines_to_reformat).split()
+
+        # And reshape.
+        width = (buffer.text_width or 80) - len(indent)
+        reshaped_text = [indent]
+        current_width = 0
+        for w in words:
+            if current_width:
+                if len(w) + current_width + 1 > width:
+                    reshaped_text.append('\n')
+                    reshaped_text.append(indent)
+                    current_width = 0
+                else:
+                    reshaped_text.append(' ')
+                    current_width += 1
+
+            reshaped_text.append(w)
+            current_width += len(w)
+
+        if reshaped_text[-1] != '\n':
+            reshaped_text.append('\n')
+
+        # Apply result.
+        buffer.document = Document(
+            text=''.join(lines_before + reshaped_text + lines_after),
+            cursor_position=len(''.join(lines_before + reshaped_text)))
