@@ -32,6 +32,8 @@ if six.PY2:
 else:
     ascii_lowercase = string.ascii_lowercase
 
+vi_register_names = ascii_lowercase + '0123456789'
+
 
 class TextObjectType(object):
     EXCLUSIVE = 'EXCLUSIVE'
@@ -209,26 +211,40 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         """
         event.current_buffer.cursor_down(count=event.arg)
 
-    @handle('k', filter=navigation_mode)
     @handle(Keys.Up, filter=navigation_mode)
     @handle(Keys.ControlP, filter=navigation_mode)
     def _(event):
         """
         Arrow up and ControlP in navigation mode go up.
         """
-        b = event.current_buffer
-        b.auto_up(count=event.arg)
+        event.current_buffer.auto_up(count=event.arg)
+
+    @handle('k', filter=navigation_mode)
+    def _(event):
+        """
+        Go up, but if we enter a new history entry, move to the start of the
+        line.
+        """
+        event.current_buffer.auto_up(
+            count=event.arg, go_to_start_of_line_if_history_changes=True)
 
-    @handle('j', filter=navigation_mode)
     @handle(Keys.Down, filter=navigation_mode)
     @handle(Keys.ControlN, filter=navigation_mode)
     def _(event):
         """
         Arrow down and Control-N in navigation mode.
         """
-        b = event.current_buffer
-        b.auto_down(count=event.arg)
+        event.current_buffer.auto_down(count=event.arg)
 
+    @handle('j', filter=navigation_mode)
+    def _(event):
+        """
+        Go down, but if we enter a new history entry, go to the start of the line.
+        """
+        event.current_buffer.auto_down(
+            count=event.arg, go_to_start_of_line_if_history_changes=True)
+
+    @handle(Keys.ControlH, filter=navigation_mode)
     @handle(Keys.Backspace, filter=navigation_mode)
     def _(event):
         """
@@ -356,7 +372,8 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         lines = buffer.document.lines
 
         before = '\n'.join(lines[:buffer.document.cursor_position_row])
-        deleted = '\n'.join(lines[buffer.document.cursor_position_row: buffer.document.cursor_position_row + event.arg])
+        deleted = '\n'.join(lines[buffer.document.cursor_position_row:
+                                  buffer.document.cursor_position_row + event.arg])
         after = '\n'.join(lines[buffer.document.cursor_position_row + event.arg:])
 
         # Set new text.
@@ -393,15 +410,26 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
 
     @handle('J', filter=navigation_mode & ~IsReadOnly())
     def _(event):
-        """ Join lines. """
+        " Join lines. "
         for i in range(event.arg):
             event.current_buffer.join_next_line()
 
+    @handle('g', 'J', filter=navigation_mode & ~IsReadOnly())
+    def _(event):
+        " Join lines without space. "
+        for i in range(event.arg):
+            event.current_buffer.join_next_line(separator='')
+
     @handle('J', filter=selection_mode & ~IsReadOnly())
     def _(event):
-        """ Join selected lines. """
+        " Join selected lines. "
         event.current_buffer.join_selected_lines()
 
+    @handle('g', 'J', filter=selection_mode & ~IsReadOnly())
+    def _(event):
+        " Join selected lines without space. "
+        event.current_buffer.join_selected_lines(separator='')
+
     @handle('n', filter=navigation_mode)
     def _(event):  # XXX: use `text_object`
         """
@@ -439,26 +467,25 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
             before=True,
             count=event.arg)
 
-    def paste_named_register_handler(c):
-        @handle('"', c, 'p')
-        def _(event):
-            " Paste from named register. "
+    @handle('"', Keys.Any, 'p', filter=navigation_mode)
+    def _(event):
+        " Paste from named register. "
+        c = event.key_sequence[1].data
+        if c in vi_register_names:
             data = event.cli.vi_state.named_registers.get(c)
             if data:
                 event.current_buffer.paste_clipboard_data(data, count=event.arg)
 
-        @handle('"', c, 'P')
-        def _(event):
-            " Paste (before) from named register. "
+    @handle('"', Keys.Any, 'P', filter=navigation_mode)
+    def _(event):
+        " Paste (before) from named register. "
+        c = event.key_sequence[1].data
+        if c in vi_register_names:
             data = event.cli.vi_state.named_registers.get(c)
             if data:
                 event.current_buffer.paste_clipboard_data(
                     data, before=True, count=event.arg)
 
-    for c in ascii_lowercase + '0123456789':
-        paste_named_register_handler(c)
-
-
     @handle('r', Keys.Any, filter=navigation_mode)
     def _(event):
         """
@@ -641,8 +668,25 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         c = buffer.document.current_char
 
         if c is not None and c != '\n':
-            c = (c.upper() if c.islower() else c.lower())
-            buffer.insert_text(c, overwrite=True)
+            buffer.insert_text(c.swapcase(), overwrite=True)
+
+    @handle('g', 'u', 'u', filter=navigation_mode & ~IsReadOnly())
+    def _(event):
+        " Lowercase current line. "
+        buff = event.current_buffer
+        buff.transform_current_line(lambda s: s.lower())
+
+    @handle('g', 'U', 'U', filter=navigation_mode & ~IsReadOnly())
+    def _(event):
+        " Uppercase current line. "
+        buff = event.current_buffer
+        buff.transform_current_line(lambda s: s.upper())
+
+    @handle('g', '~', '~', filter=navigation_mode & ~IsReadOnly())
+    def _(event):
+        " Swap case of the current line. "
+        buff = event.current_buffer
+        buff.transform_current_line(lambda s: s.swapcase())
 
     @handle('#', filter=navigation_mode)
     def _(event):
@@ -825,15 +869,15 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
     # *** Operators ***
     #
 
-    def create_delete_and_change_operators(delete_only, reg_name=None):
+    def create_delete_and_change_operators(delete_only, with_register=False):
         """
         Delete and change operators.
 
         :param delete_only: Create an operator that deletes, but doesn't go to insert mode.
-        :param reg_name: Copy the deleted text to this named register instead of the clipboard.
+        :param with_register: Copy the deleted text to this named register instead of the clipboard.
         """
-        if reg_name:
-            handler_keys = ('"', reg_name, 'cd'[delete_only])
+        if with_register:
+            handler_keys = ('"', Keys.Any, 'cd'[delete_only])
         else:
             handler_keys = 'cd'[delete_only]
 
@@ -848,8 +892,10 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
 
             # Set deleted/changed text to clipboard or named register.
             if clipboard_data and clipboard_data.text:
-                if reg_name:
-                    event.cli.named_registers[reg_name] = clipboard_data
+                if with_register:
+                    reg_name = event.key_sequence[1].data
+                    if reg_name in vi_register_names:
+                        event.cli.vi_state.named_registers[reg_name] = clipboard_data
                 else:
                     event.cli.clipboard.set_data(clipboard_data)
 
@@ -857,9 +903,10 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
             if not delete_only:
                 event.cli.vi_state.input_mode = InputMode.INSERT
 
-    for c in tuple(ascii_lowercase) + (None, ):
-        create_delete_and_change_operators(False, reg_name=c)
-        create_delete_and_change_operators(True, reg_name=c)
+    create_delete_and_change_operators(False, False)
+    create_delete_and_change_operators(False, True)
+    create_delete_and_change_operators(True, False)
+    create_delete_and_change_operators(True, True)
 
     def create_transform_handler(filter, transform_func, *a):
         @operator(*a, filter=filter & ~IsReadOnly())
@@ -892,16 +939,14 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         if clipboard_data.text:
             event.cli.clipboard.set_data(clipboard_data)
 
-    def create_yank_selection_handler(c):
-        @operator('"', c, 'y')
-        def _(event, text_object):
-            " Yank selection to named register 'c'. "
+    @operator('"', Keys.Any, 'y')
+    def _(event, text_object):
+        " Yank selection to named register. "
+        c = event.key_sequence[1].data
+        if c in vi_register_names:
             _, clipboard_data = text_object.cut(event.current_buffer)
             event.cli.vi_state.named_registers[c] = clipboard_data
 
-    for c in ascii_lowercase + '0123456789':
-        create_yank_selection_handler(c)
-
     @operator('>')
     def _(event, text_object):
         """
@@ -1018,16 +1063,13 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         """
         return TextObject(event.current_buffer.document.get_start_of_line_position(after_whitespace=False))
 
-    def create_ci_ca_handles(ci_start, ci_end, inner):
-                # TODO: 'dab', 'dib', (brackets or block) 'daB', 'diB', Braces.
+    def create_ci_ca_handles(ci_start, ci_end, inner, key=None):
                 # TODO: 'dat', 'dit', (tags (like xml)
         """
         Delete/Change string between this start and stop character. But keep these characters.
         This implements all the ci", ci<, ci{, ci(, di", di<, ca", ca<, ... combinations.
         """
-        @text_object('ai'[inner], ci_start, no_move_handler=True)
-        @text_object('ai'[inner], ci_end, no_move_handler=True)
-        def _(event):
+        def handler(event):
             if ci_start == ci_end:
                 # Quotes
                 start = event.current_buffer.document.find_backwards(ci_start, in_current_line=False)
@@ -1044,11 +1086,20 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
                 # Nothing found.
                 return TextObject(0)
 
+        if key is None:
+            text_object('ai'[inner], ci_start, no_move_handler=True)(handler)
+            text_object('ai'[inner], ci_end, no_move_handler=True)(handler)
+        else:
+            text_object('ai'[inner], key, no_move_handler=True)(handler)
+
     for inner in (False, True):
         for ci_start, ci_end in [('"', '"'), ("'", "'"), ("`", "`"),
                                  ('[', ']'), ('<', '>'), ('{', '}'), ('(', ')')]:
             create_ci_ca_handles(ci_start, ci_end, inner)
 
+        create_ci_ca_handles('(', ')', inner, 'b')  # 'dab', 'dib'
+        create_ci_ca_handles('{', '}', inner, 'B')  # 'daB', 'diB'
+
     @text_object('{')
     def _(event):
         """
@@ -1448,6 +1499,8 @@ def load_vi_bindings(registry, enable_visual_key=Always(),
         try:
             # Lookup.
             code = (event.cli.vi_state.digraph_symbol1, event.data)
+            if code not in DIGRAPHS:
+                code = code[::-1]  # Try reversing.
             symbol = DIGRAPHS[code]
         except KeyError:
             # Unkown digraph.
@@ -1603,6 +1656,7 @@ def load_vi_search_bindings(registry, get_search_state=None,
 
     @handle(Keys.Escape, filter=has_focus)
     @handle(Keys.ControlC, filter=has_focus)
+    @handle(Keys.ControlH, filter=has_focus & Condition(search_buffer_is_empty))
     @handle(Keys.Backspace, filter=has_focus & Condition(search_buffer_is_empty))
     def _(event):
         """
