@@ -4,13 +4,15 @@ Renders the command line on the console.
 """
 from __future__ import unicode_literals
 
-from pygments.style import Style
-from pygments.token import Token
-from prompt_toolkit.layout.screen import Point, Screen, WritePosition
-from prompt_toolkit.layout.mouse_handlers import MouseHandlers
-from prompt_toolkit.output import Output
-from prompt_toolkit.utils import is_windows
 from prompt_toolkit.filters import to_cli_filter
+from prompt_toolkit.layout.mouse_handlers import MouseHandlers
+from prompt_toolkit.layout.screen import Point, Screen, WritePosition
+from prompt_toolkit.output import Output
+from prompt_toolkit.styles import Style
+from prompt_toolkit.token import Token
+from prompt_toolkit.utils import is_windows
+
+from six.moves import range
 
 __all__ = (
     'Renderer',
@@ -18,37 +20,36 @@ __all__ = (
 )
 
 
-class _StyleForTokenCache(dict):
+def _output_screen_diff(output, screen, current_pos, previous_screen=None, last_char=None,
+                        is_done=False, attrs_for_token=None, size=None, previous_width=0):  # XXX: drop is_done
     """
-    A cache structure that maps Pygments Tokens to their style objects.
-    """
-    def __init__(self, style):
-        self.style = style
+    Render the diff between this screen and the previous screen.
 
-    def __missing__(self, token):
-        try:
-            result = self.style.style_for_token(token)
-        except KeyError:
-            result = None
-
-        self[token] = result
-        return result
-
-
-def output_screen_diff(output, screen, current_pos, previous_screen=None, last_char=None,
-                       is_done=False, style=None):  # XXX: drop is_done
-    """
-    Create diff of this screen with the previous screen.
+    This takes two `Screen` instances. The one that represents the output like
+    it was during the last rendering and one that represents the current
+    output raster. Looking at these two `Screen` instances, this function will
+    render the difference by calling the appropriate methods of the `Output`
+    object that only paint the changes to the terminal.
 
     This is some performance-critical code which is heavily optimized.
     Don't change things without profiling first.
+
+    :param current_pos: Current cursor position.
+    :param last_char: `Char` instance that represents the output attributes of
+            the last drawn character. (Color/attributes.)
+    :param attrs_for_token: :class:`._TokenToAttrsCache` instance.
+    :param width: The width of the terminal.
+    :param prevous_width: The width of the terminal during the last rendering.
     """
+    width, height = size.columns, size.rows
+
     #: Remember the last printed character.
     last_char = [last_char]  # nonlocal
     background_turned_on = [False]  # Nonlocal
 
     #: Variable for capturing the output.
     write = output.write
+    write_raw = output.write_raw
 
     # Create locals for the most used output methods.
     # (Save expensive attribute lookups.)
@@ -58,6 +59,9 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
     _output_cursor_up = output.cursor_up
     _output_cursor_backward = output.cursor_backward
 
+    # Hide cursor before rendering. (Avoid flickering.)
+    output.hide_cursor()
+
     def reset_attributes():
         " Wrapper around Output.reset_attributes. "
         _output_reset_attributes()
@@ -80,18 +84,16 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
         elif new.y < current_y:
             _output_cursor_up(current_y - new.y)
 
-        if current_x >= screen.width - 1:
+        if current_x >= width - 1:
             write('\r')
             _output_cursor_forward(new.x)
-        elif new.x < current_x or current_x >= screen.width - 1:
+        elif new.x < current_x or current_x >= width - 1:
             _output_cursor_backward(current_x - new.x)
         elif new.x > current_x:
             _output_cursor_forward(new.x - current_x)
 
         return new
 
-    style_for_token = _StyleForTokenCache(style)
-
     def output_char(char):
         """
         Write the output of this character.
@@ -101,18 +103,12 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
         if last_char[0] and last_char[0].token == char.token:
             write(char.char)
         else:
-            style = style_for_token[char.token]
+            attrs = attrs_for_token[char.token]
 
-            if style:
-                _output_set_attributes(style['color'], style['bgcolor'],
-                                       bold=style.get('bold', False),
-                                       underline=style.get('underline', False))
+            _output_set_attributes(attrs)
 
-                # If we print something with a background color, remember that.
-                background_turned_on[0] = bool(style['bgcolor'])
-            else:
-                # Reset previous style and output.
-                reset_attributes()
+            # If we print something with a background color, remember that.
+            background_turned_on[0] = bool(attrs.bgcolor)
 
             write(char.char)
 
@@ -125,7 +121,7 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
 
     # When the previous screen has a different size, redraw everything anyway.
     # Also when we are done. (We meight take up less rows, so clearing is important.)
-    if is_done or not previous_screen or previous_screen.width != screen.width:  # XXX: also consider height??
+    if is_done or not previous_screen or previous_width != width:  # XXX: also consider height??
         current_pos = move_cursor(Point(0, 0))
         reset_attributes()
         output.erase_down()
@@ -133,19 +129,21 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
         previous_screen = Screen()
 
     # Get height of the screen.
-    # (height changes as we loop over _buffer, so remember the current value.)
-    current_height = screen.height
+    # (height changes as we loop over data_buffer, so remember the current value.)
+    # (Also make sure to clip the height to the size of the output.)
+    current_height = min(screen.height, height)
 
     # Loop over the rows.
-    row_count = max(screen.height, previous_screen.height)
+    row_count = min(max(screen.height, previous_screen.height), height)
     c = 0  # Column counter.
 
-    for y, r in enumerate(range(0, row_count)):
-        new_row = screen._buffer[r]
-        previous_row = previous_screen._buffer[r]
+    for y in range(row_count):
+        new_row = screen.data_buffer[y]
+        previous_row = previous_screen.data_buffer[y]
+        zero_width_escapes_row = screen.zero_width_escapes[y]
 
-        new_max_line_len = max(new_row.keys()) if new_row else 0
-        previous_max_line_len = max(previous_row.keys()) if previous_row else 0
+        new_max_line_len = min(width - 1, max(new_row.keys()) if new_row else 0)
+        previous_max_line_len = min(width - 1, max(previous_row.keys()) if previous_row else 0)
 
         # Loop over the columns.
         c = 0
@@ -159,12 +157,17 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
             # `Char.__ne__`, but inline the same expression.)
             if new_char.char != old_char.char or new_char.token != old_char.token:
                 current_pos = move_cursor(Point(y=y, x=c))
+
+                # Send injected escape sequences to output.
+                if c in zero_width_escapes_row:
+                    write_raw(zero_width_escapes_row[c])
+
                 output_char(new_char)
                 current_pos = current_pos._replace(x=current_pos.x + char_width)
 
             c += char_width
 
-        # If the new line is shorter, trim it
+        # If the new line is shorter, trim it.
         if previous_screen and new_max_line_len < previous_max_line_len:
             current_pos = move_cursor(Point(y=y, x=new_max_line_len+1))
             reset_attributes()
@@ -180,8 +183,8 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
     # the artifact of the input scrolling when the completion menu is shown.
     # (If the scrolling is actually wanted, the layout can still be build in a
     # way to behave that way by setting a dynamic height.)
-    if screen.height > previous_screen.height:
-        current_pos = move_cursor(Point(y=screen.height - 1, x=0))
+    if current_height > previous_screen.height:
+        current_pos = move_cursor(Point(y=current_height - 1, x=0))
 
     # Move cursor:
     if is_done:
@@ -200,6 +203,9 @@ def output_screen_diff(output, screen, current_pos, previous_screen=None, last_c
     if background_turned_on[0]:
         reset_attributes()
 
+    if screen.show_cursor or is_done:
+        output.show_cursor()
+
     return current_pos, last_char[0]
 
 
@@ -207,6 +213,24 @@ class HeightIsUnknownError(Exception):
     " Information unavailable. Did not yet receive the CPR response. "
 
 
+class _TokenToAttrsCache(dict):
+    """
+    A cache structure that maps Pygments Tokens to :class:`.Attr`.
+    (This is an important speed up.)
+    """
+    def __init__(self, get_style_for_token):
+        self.get_style_for_token = get_style_for_token
+
+    def __missing__(self, token):
+        try:
+            result = self.get_style_for_token(token)
+        except KeyError:
+            result = None
+
+        self[token] = result
+        return result
+
+
 class Renderer(object):
     """
     Typical usage:
@@ -214,22 +238,29 @@ class Renderer(object):
     ::
 
         output = Vt100_Output.from_pty(sys.stdout)
-        r = Renderer(output)
-        r.render(cli, layout=..., style=...)
+        r = Renderer(style, output)
+        r.render(cli, layout=...)
     """
-    def __init__(self, output, use_alternate_screen=False, mouse_support=False):
+    def __init__(self, style, output, use_alternate_screen=False, mouse_support=False):
+        assert isinstance(style, Style)
         assert isinstance(output, Output)
 
+        self.style = style
         self.output = output
         self.use_alternate_screen = use_alternate_screen
         self.mouse_support = to_cli_filter(mouse_support)
 
         self._in_alternate_screen = False
         self._mouse_support_enabled = False
+        self._bracketed_paste_enabled = False
+
+        # Waiting for CPR flag. True when we send the request, but didn't got a
+        # response.
+        self.waiting_for_cpr = False
 
         self.reset(_scroll=True)
 
-    def reset(self, _scroll=False):
+    def reset(self, _scroll=False, leave_alternate_screen=True):
         # Reset position
         self._cursor_pos = Point(x=0, y=0)
 
@@ -240,8 +271,11 @@ class Renderer(object):
         self._last_screen = None
         self._last_size = None
         self._last_char = None
-        self._last_style = None  # When the style changes, we have to do a full
-                                 # redraw as well.
+
+        # When the style hash changes, we have to do a full redraw as well as
+        # clear the `_attrs_for_token` dictionary.
+        self._last_style_hash = None
+        self._attrs_for_token = None
 
         # Default MouseHandlers. (Just empty.)
         self.mouse_handlers = MouseHandlers()
@@ -259,9 +293,8 @@ class Renderer(object):
             self.output.scroll_buffer_to_prompt()
 
         # Quit alternate screen.
-        if self._in_alternate_screen:
+        if self._in_alternate_screen and leave_alternate_screen:
             self.output.quit_alternate_screen()
-            self.output.flush()
             self._in_alternate_screen = False
 
         # Disable mouse support.
@@ -269,6 +302,11 @@ class Renderer(object):
             self.output.disable_mouse_support()
             self._mouse_support_enabled = False
 
+        # Disable bracketed paste.
+        if self._bracketed_paste_enabled:
+            self.output.disable_bracketed_paste()
+            self._bracketed_paste_enabled = False
+
         # Flush output. `disable_mouse_support` needs to write to stdout.
         self.output.flush()
 
@@ -315,6 +353,7 @@ class Renderer(object):
                 self._min_available_height = self.output.get_size().rows
             else:
                 # Asks for a cursor position report (CPR).
+                self.waiting_for_cpr = True
                 self.output.ask_for_cpr()
 
     def report_absolute_cursor_row(self, row):
@@ -330,14 +369,15 @@ class Renderer(object):
         # Set the
         self._min_available_height = rows_below_cursor
 
-    def render(self, cli, layout, style=None, is_done=False):
+        self.waiting_for_cpr = False
+
+    def render(self, cli, layout, is_done=False):
         """
         Render the current interface to the output.
 
         :param is_done: When True, put the cursor at the end of the interface. We
                 won't print any changes to this part.
         """
-        style = style or Style
         output = self.output
 
         # Enter alternate screen.
@@ -345,6 +385,11 @@ class Renderer(object):
             self._in_alternate_screen = True
             output.enter_alternate_screen()
 
+        # Enable bracketed paste.
+        if not self._bracketed_paste_enabled:
+            self.output.enable_bracketed_paste()
+            self._bracketed_paste_enabled = True
+
         # Enable/disable mouse support.
         needs_mouse_support = self.mouse_support(cli)
 
@@ -359,6 +404,8 @@ class Renderer(object):
         # Create screen and write layout to it.
         size = output.get_size()
         screen = Screen()
+        screen.show_cursor = False  # Hide cursor by default, unless one of the
+                                    # containers decides to display it.
         mouse_handlers = MouseHandlers()
 
         if is_done:
@@ -374,9 +421,12 @@ class Renderer(object):
         # When we render using another style, do a full repaint. (Forget about
         # the previous rendered screen.)
         # (But note that we still use _last_screen to calculate the height.)
-        if style != self._last_style:
+        if self.style.invalidation_hash() != self._last_style_hash:
             self._last_screen = None
-        self._last_style = style
+            self._attrs_for_token = None
+        if self._attrs_for_token is None:
+            self._attrs_for_token = _TokenToAttrsCache(self.style.get_attrs_for_token)
+        self._last_style_hash = self.style.invalidation_hash()
 
         layout.write_to_screen(cli, screen, mouse_handlers, WritePosition(
             xpos=0,
@@ -391,11 +441,12 @@ class Renderer(object):
             screen.replace_all_tokens(Token.Aborted)
 
         # Process diff and write to output.
-        self._cursor_pos, self._last_char = output_screen_diff(
+        self._cursor_pos, self._last_char = _output_screen_diff(
             output, screen, self._cursor_pos,
             self._last_screen, self._last_char, is_done,
-            style=style,
-            )
+            attrs_for_token=self._attrs_for_token,
+            size=size,
+            previous_width=(self._last_size.columns if self._last_size else 0))
         self._last_screen = screen
         self._last_size = size
         self.mouse_handlers = mouse_handlers
@@ -412,11 +463,15 @@ class Renderer(object):
 
         output.flush()
 
-    def erase(self):
+    def erase(self, leave_alternate_screen=True, erase_title=True):
         """
         Hide all output and put the cursor back at the first line. This is for
         instance used for running a system command (while hiding the CLI) and
         later resuming the same CLI.)
+
+        :param leave_alternate_screen: When True, and when inside an alternate
+            screen buffer, quit the alternate screen.
+        :param erase_title: When True, clear the title from the title bar.
         """
         output = self.output
 
@@ -427,10 +482,10 @@ class Renderer(object):
         output.flush()
 
         # Erase title.
-        if self._last_title:
+        if self._last_title and erase_title:
             output.clear_title()
 
-        self.reset()
+        self.reset(leave_alternate_screen=leave_alternate_screen)
 
     def clear(self):
         """
@@ -454,22 +509,20 @@ def print_tokens(output, tokens, style):
     Print a list of (Token, text) tuples in the given style to the output.
     """
     assert isinstance(output, Output)
-    assert style
+    assert isinstance(style, Style)
 
     # Reset first.
     output.reset_attributes()
     output.enable_autowrap()
 
     # Print all (token, text) tuples.
-    style_for_token = _StyleForTokenCache(style)
+    attrs_for_token = _TokenToAttrsCache(style.get_attrs_for_token)
 
     for token, text in tokens:
-        style = style_for_token[token]
+        attrs = attrs_for_token[token]
 
-        if style:
-            output.set_attributes(style['color'], style['bgcolor'],
-                                  bold=style.get('bold', False),
-                                  underline=style.get('underline', False))
+        if attrs:
+            output.set_attributes(attrs)
         else:
             output.reset_attributes()
 
